
;;; query-sheet.el --- multiple queries across multiple files 

;; Keywords: convenience, searching, editing

;; This file is not part of GNU Emacs.
;; 
;; query-sheet.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; query-sheet.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.


;;; Commentary: 
;;-------------
::
;; installation: put this file in your path and evaluate (require 'query-sheet)
;;
;;    ************************************************************************
;;    WARNING: DO NOT BYTE-COMPILE THIS CODE !
;;             the unusual usage of defmacro in qsheet-query does not allow it
;;             to run properly when byte-compiled (except if :wrapper-macro is 
;;             nil for all modes you use)
;;    ************************************************************************
;;
;; usage:
;;
;;    the main function is 'qsheet-query :
;;
;;    (qsheet-query SHEET MODE &optional STRING)
;;    writes in the buffer associated to SHEET the result of the query MODE on
;;    argument STRING
;;
;;    all the interesting job is done when defining SHEET and MODE
;;
;;    SHEET is a querysheet structure returned by
;;    (make-querysheet 
;;      :buffer-name    ...a string, the name of the associated buffer
;;	:sources        ...a function of no argument
;;                         returning a list of file names
;;	:modes          ...a list of modes (see below)
;;    )
;;
;;    MODE is a querymode structure returned by       
;;    (make-querymode 
;;      :label        ...a string describing the query (used to label a button)
;;      :what-i-did   ...a string used to recall the query action (for information)
;;      :case         ...a boolean (t or nil) specifying case-fold-search 
;;      :interactive  ...a function of no argument used when no STRING
;;                       argument is given to 'qsheet-query and returning STRING
;;      :get      ...a function of two arguments STRING and SOURCE
;;                   returning the RESULT of the next query for STRING in file SOURCE
;;                   and returning nil if there is no match.
;;                   (it will be invoked as many times as necessary to get all matches
;;                   for STRING in SOURCE; so it should take the point as is, and 
;;                   it should move the point forward across the match) 
;;                   --> REMEMBER: this code is evaluated while in the source buffer
;;      :display  ...a function of one argument RESULT (as returned by :get)
;;                   which insert some information in the SHEET buffer
;;                   (there is no need to set the buffer or the point, this is
;;                   done automatically) 
;;                   --> REMEMBER: this code is evaluated while in the display buffer
;;      :wrapper-macro  ...an optional quoted macro body wrapped around the whole 
;;                         query operation ,@body which can be used to define
;;                         local variables used by :get and :display allowing for
;;                         example the computing of statistics on the query result
;;                         --> IMPORTANT: the code around ,@body is evaluated in 
;;                         the display buffer (the query sheet)
;;    )
;;
;;
;;    ok, this is a quite terse and obscure description. 
;;    ...so please check the detailed example provided just below
;;  
;;                                                      Stef <hepta@zogotounga.net>
;;
;;
;; last version:    http://www.zogotounga.net/comp/index.html
;; ...also within the keykit-mode package:  http://www.zogotounga.net/GM/GMemacs.html
;;
;;
;; last modified May 12, 2002
;; (requires Emacs 21.1)



;;; ===================================================== basic example:
;;;
;;; try (qsheet-example)
;;; and have a look at the code:

(defun qsheet-example ()
  "an example of query-sheet usage. we define a sheet exploring this very file, 
query-sheet.el, assuming it is located in your ~/site-lisp/ directory. we also
define specific modes by copying and sligthly customizing the provided basic modes.
much more interesting changes are possible by setf-ing the :wrapper-macro, :get 
and :display slots of the modes
...see the CL info node for more details about working with defstruct and structures"

  ;; don't forget this !
  (require 'query-sheet)

  ;; here we define a wrapper for the basic functionality:
  (defun qsheet-example-search (&optional string)
    "an example of 'qsheet-query usage"
    (interactive)
    (qsheet-query qsheet-example-sheet qsheet-example-query string))

  ;; this defines the sheet configuration:
  (defvar qsheet-example-sheet
    (make-querysheet :buffer-name "*Example Query Sheet*"
		     :sources (lambda () '("~/site-lisp/query-sheet.el"))
		     :modes '(qsheet-example-query
			      qsheet-example-regexp-query)))

  ;; this defines the first mode (search for a string):
  ;; (see the definition of qsheet-search-query at the end of this file)
  (defvar qsheet-example-query 
    (copy-querymode qsheet-search-query))
  (setf (querymode-label qsheet-example-query)
	"Look in ~/site-lisp/query-sheet.el for a string")

  ;; this defines the second mode (search for a regexp):
  ;; (see the definition of qsheet-search-regexp-query at the end of this file)
  (defvar qsheet-example-regexp-query 
    (copy-querymode qsheet-search-regexp-query))
  (setf (querymode-label qsheet-example-regexp-query)
	"Search ~/site-lisp/query-sheet.el for a regexp")
  (setf (querymode-case qsheet-example-regexp-query) nil)

  ;; now let's go:

  (qsheet-example-search "label")

  ;;... you should get a list of  all occurences of the word "label" in this file
  ;;    choose one, then click in the --> corresponding location (Line ??) to jump there
)

;;; ======================================================= (end of example)


;;; -----------------------------------------------------------------------------
;;;
;;;  core code
;;;
;;; -----------------------------------------------------------------------------


(defstruct querysheet buffer-name sources modes)
(defstruct querymode label what-i-did case interactive get display wrapper-macro)


(defun qsheet-query (sheet mode &optional string)
  "Go through all the files in SHEET sources and returns a clickable list of
all occurence of STRING found according to MODE in SHEET buffer."
  (interactive)
  (setq string (or string 
		   (funcall (querymode-interactive mode)))) 
  (let ((sources (funcall (querysheet-sources sheet))))
    (eval `(defmacro qsheet-query-wrap-macro (&rest body)
	     ,(or (querymode-wrapper-macro mode)
		  '`(progn ,@body))))
    (qsheet-insert-search-headers string mode sheet)
    (setq buffer-read-only nil)
    (save-excursion
      (qsheet-query-wrap-macro
       (dolist (source sources)
	 (if (file-readable-p source)
	     (with-temp-buffer
	       (setq case-fold-search (querymode-case mode))
	       (insert-file-contents source)
	       (beginning-of-buffer)
	       (let (results)
		 (while (setq results (funcall (querymode-get mode) string source))
		   (save-excursion
		     (set-buffer (qsheet-get-buffer-create-sheet sheet))
		     (funcall (querymode-display mode) results))))))))
      (goto-char (point-max))
      (insert-char ?\n 3)
      (setq buffer-read-only t))
    (pop-to-buffer (qsheet-get-buffer-create-sheet sheet))))


(defun qsheet-get-buffer-create-sheet (sheet)
  "get-buffer-create SHEET :buffer-name, inserting the header if necessary"
  (let ((buff (get-buffer (querysheet-buffer-name sheet))))
    (unless buff
	(save-excursion
	  (setq buff (get-buffer-create (querysheet-buffer-name sheet)))
	  (set-buffer buff)
          (qsheet-insert-button "(bury)" 'bury-buffer " ")
          (qsheet-insert-button "(kill)" '(lambda () (interactive) (kill-buffer ())) " ")
          (qsheet-insert-button "(reset)" `(lambda () 
					     (interactive)
					     (kill-buffer ()) 
					     (switch-to-buffer
					      (qsheet-get-buffer-create-sheet ,sheet)))
				"\n")
	  (dolist (mode (querysheet-modes sheet))
	    (qsheet-insert-button (querymode-label (eval mode))
				  `(lambda ()
				    (interactive)
				    (qsheet-query ,sheet ,(eval mode)))
				  "\n"))  
	  (insert "\n\n")))
    buff))


(defun qsheet-insert-search-headers (keyword mode sheet)
  "insert a couple of buttons at the beginning of the query results"
  (set-buffer (qsheet-get-buffer-create-sheet sheet))
  (setq buffer-read-only nil)
  (goto-char (point-max))
  (insert (propertize (concat ">>> " (querymode-what-i-did mode) ": ")
		      'face 'bold
		      'mode mode)
	  keyword ?\n
	  (propertize "Results are:" 'face 'bold) 
	  ?\n)
  (qsheet-insert-button "Erase" 'qsheet-erase-search " ")
  (qsheet-insert-button "Refresh" `(lambda () 
				     (interactive) 
				     (qsheet-erase-search)
				     (qsheet-query ,sheet ,mode ,keyword))
			" ")
  (qsheet-insert-button "Other" `(lambda () (interactive) (qsheet-query ,sheet ,mode)) " ")
  (qsheet-insert-button "Case" 
			`(lambda ()
			  (interactive)
			  (if (y-or-n-p (concat 
					 "Case is currently "
					 (if (querymode-case ,mode) "ignored" "respected")
					 ". Toggle ? "))
			      (setf (querymode-case ,mode) (null (querymode-case ,mode))))))
  (insert "(was "
	  (if (querymode-case mode) "ignored" "respected")
	  ")")
  (insert-char ?\n 2)
  (setq buffer-read-only t))


(defun qsheet-erase-search ()
  "used by the 'Erase' and 'Refresh' buttons"
  (interactive)
  (previous-line 2)
  (beginning-of-line)
  (setq buffer-read-only nil)
  (delete-region (point) (save-excursion (next-line 4)
					 (if (null (re-search-forward "^>>> " nil t))
					     (goto-char (point-max))
					   (match-beginning 0))))
  (setq buffer-read-only t))


(defun qsheet-insert-button (text action &optional delimiter)
  "insert a button labelled TEXT which triggers ACTION when clicked"
  (insert (propertize (concat " " text " ")
		      'face 'custom-button-face
		      'mouse-face 'custom-button-pressed-face
		      'local-map (qsheet-local-map action)))
  (if delimiter
      (insert delimiter)))


(defun qsheet-local-map (action)
  "return a local map associating ACTION to mouse clicks, <RETURN> and <SPACE>"
  (let ((map (make-sparse-keymap)))
    (define-key map [mouse-1] action)
    (define-key map [mouse-2] action)
    (define-key map "\r" action)
    (define-key map " " action)
    map))


;;; -----------------------------------------------------------------------------
;;;
;;; defining basic functionality, for further copy/customize:
;;;
;;; -----------------------------------------------------------------------------

(defvar qsheet-blank-sheet
  (make-querysheet :buffer-name "*Query Sheet*"
		   :sources 'ignore
		   :modes '(qsheet-search-query
			    qsheet-search-word-query
			    qsheet-search-exact-word-query
			    qsheet-search-regexp-query))
  "a query sheet templates featuring the four basic query modes.
it is not associated to any source")


(defun new-querysheet (&optional sources name modes)
  "return a copy of 'qsheet-blank-sheet, customizable through the optional arguments
the provided modes are a plain search and a regexp search
...no sources are defined by default"
  (let ((qsheet (copy-querysheet qsheet-blank-sheet)))
    (when sources
      (setf (querysheet-sources qsheet) sources)) 
    (when name
      (setf (querysheet-buffer-name qsheet) name)) 
    (when modes
      (setf (querysheet-modes qsheet) modes))
    qsheet))


(defun qsheet-make-query-functions (sheet prefix)
  "define a set of functions that can be used interactively to invoque the query modes
associated to SHEET.
the function names are derived from the mode names, with PREFIX replacing \"qsheet-\",
or PREFIX directly prepended if the mode name do not start with \"qsheet-\""
  (dolist (mode (querysheet-modes sheet))
    (let ((function-name (if (and (> (length (symbol-name mode)) 7)
			     (string= "qsheet-" (substring (symbol-name mode) 0 7)))
			(concat prefix (substring (symbol-name mode) 7 ))
		      (concat prefix (symbol-name mode)))))
      (unless (eval `(functionp (quote ,(read function-name))))
	(eval `(defun ,(read function-name)
		 (&optional string)
		 (interactive)
		 (qsheet-query ,sheet ,mode string)))))))


(defvar qsheet-search-query
  (make-querymode 
   :label "Look for"
   :what-i-did "Searched for"
   :case t
   :interactive (lambda () (read-from-minibuffer "Search: " (word-at-point)))
   :wrapper-macro 
   '`(let ((counter 0)) 
      ,@body
      (insert "\n" (number-to-string counter) " matches found\n"))
   :get
   (lambda (string source) 
     (when (search-forward string nil t)
       (incf counter)
       (append (list :source source)
	       (list :found (propertize (match-string 0)  
					'face 'font-lock-warning-face))
	       (list :in-line (propertize (thing-at-point 'line)
					  'face 'font-lock-string-face))
	       (list :line-number (count-lines (point-min) (point))))))
   :display
   (lambda (result)
     (insert (plist-get result :in-line)
	     (plist-get result :found)
	     " --> "
	     (propertize (file-relative-name (plist-get result :source) "~")
			 'mouse-face 'highlight
			 'local-map (qsheet-local-map
				     `(lambda () 
					(interactive)
					(find-file ,(plist-get result :source))
					(goto-line ,(plist-get result :line-number)))))
	     " (Line " (number-to-string (plist-get result :line-number)) ")\n")))
  "a very general purpose query mode simply searching for a string in the sources.
provides a good template for further specializations such as qsheet-search-word-query")


(defvar qsheet-search-word-query
  (let ((mode (copy-querymode qsheet-search-query)))
    (setf (querymode-label mode) "Look for words containing...")
    (setf (querymode-what-i-did mode) "Looked for words containing")
    (setf (querymode-interactive mode) 
	  (lambda () (read-from-minibuffer "Word fragment: " (word-at-point))))
    (setf (querymode-get mode)    
	  (lambda (string source)
	    (when (search-forward string nil t)
	      (let ((word-found (save-excursion (backward-char 1) (word-at-point))))
		(when word-found
		  (incf counter)
		  (append (list :source source)
			  (list :found (propertize word-found	  
						   'face (if (string= word-found string) 
							     'font-lock-warning-face
							   'font-lock-keyword-face)))
			  (list :in-line (propertize (thing-at-point 'line)
				     'face 'font-lock-string-face))
			  (list :line-number
				(count-lines (point-min) (point)))))))))
    mode)
  "a query mode derived from 'qsheet-search-query and searching for words embedding a string")


(defvar qsheet-search-exact-word-query
  (let ((mode (copy-querymode qsheet-search-query)))
    (setf (querymode-label mode) "Look for a word")
    (setf (querymode-what-i-did mode) "Looked for word")
    (setf (querymode-interactive mode) 
	  (lambda () (read-from-minibuffer "Word: " (word-at-point))))
    (setf (querymode-get mode)    
	  (lambda (string source)
	    (when (search-forward string nil t)
	      (let ((word-found (save-excursion (backward-char 1) (word-at-point))))
		(when (and word-found (= (length word-found) (length string)))
		  (incf counter)
		  (append (list :source source)
			  (list :found (propertize word-found
						   'face 'font-lock-warning-face ))
			  (list :in-line (propertize (thing-at-point 'line)
						     'face 'font-lock-string-face))
			  (list :line-number (count-lines (point-min) (point)))))))))
    mode)
  "a query mode derived from 'qsheet-search-query and searching for precise words")


(defvar qsheet-search-regexp-query
  (make-querymode 
   :label "Look for regexp"
   :what-i-did "Searched for regexp"
   :case t
   :interactive (lambda () (read-from-minibuffer "Search for regexp: "))
   :wrapper-macro 
   '`(let ((counter 0)) 
      ,@body
      (insert "\n" (number-to-string counter) " matches found\n"))
   :get
   (lambda (reg source) 
     (when (re-search-forward reg nil t)
       (incf counter)
       (append (list :source source)
	       (list :found (match-string 0))
	       (list :in-line (thing-at-point 'line))
	       (list :line-number (count-lines (point-min) (point))))))
   :display
   (lambda (result)
     (insert (propertize (plist-get result :in-line)
			 'face 'font-lock-string-face)
	     (propertize (plist-get result :found) 
			 'face 'bold)
	     " --> "
	     (propertize (file-relative-name (plist-get result :source) "~")
			 'mouse-face 'highlight
			 'local-map (qsheet-local-map
				     `(lambda ()
					(interactive)
					(find-file ,(plist-get result :source))
					(goto-line ,(plist-get result :line-number)))))
	     " (Line " (number-to-string (plist-get result :line-number)) ")\n")))
  "a very general purpose query mode simply searching for a regexp in the sources.
provides a good template for further specializations")


;;; -----------------------------------------------------------------------------
;;; this is it.

(provide 'query-sheet)